\
\ $ProjectVersion: 0.211 $
\
\ Code for the SPIF smartcard controller, programmable baud rate generator
\ and bq2010 interface.
\
\ Hardware: 16F873SO 26MHz
\

include picisr.fs
include libstore.fs
include libfetch.fs
include libcmove.fs
include piceeprom.fs

variable i2c-length
create i2c-buffer variable i2c-command create i2c-data $20 allot
variable i2c-addr

eecreate i2c-address 7 ee,               \ Default I2C device address

\ ----------------------------------------------------------------------
\ Version information
\ ----------------------------------------------------------------------

1 constant vsn-major
5 constant vsn-minor

\ ----------------------------------------------------------------------
\ Led
\ ----------------------------------------------------------------------

0 pin-a led
variable blinking-counter
variable blinking-speed

: led-on ( -- ) 0 blinking-speed ! led high ;
: led-off ( -- ) 0 blinking-speed ! led low ;
\ Low bit of blinking speed is set so that the test is faster in
\ interrupt routine.
:: led-blink-n ( n -- ) 1 or blinking-speed ! 1 blinking-counter ! ;
: led-blink ( -- ) 3 led-blink-n ;
: led-blink-fast ( -- ) 1 led-blink-n ;

\ ----------------------------------------------------------------------
\ Statistics and error flags
\ ----------------------------------------------------------------------

variable i2c-overflow
variable sc-parityerrors
variable sc-nack
variable sc-tocard
variable sc-fromcard

: reset-stats ( -- )
    0 i2c-overflow ! 0 sc-parityerrors ! 0 sc-nack !
    0 sc-tocard ! 0 sc-fromcard ! ;

\ ----------------------------------------------------------------------
\ Timer routines
\ ----------------------------------------------------------------------

\ With a prescaler of 8, each tick takes 1.231s

:: timer-wait ( n -- )
    tmr0 ! t0if bit-clr                    \ Setup timer 0
    begin t0if bit-set? until ;            \ Wait until overflow

: 140s ( -- ) -72 timer-wait ;
: 80ms ( -- ) $100 tmp1 v-for 0 timer-wait tmp1 v-next ;

\ The card reader frequency is 3.25MHz (26MHz/8). Each bit takes 372 clock
\ pulses, which means 114.46s, i.e. around 93 times 1.231s.
: delay-bit ( -- ) -5d timer-wait ;
: delay-half-bit ( -- ) -2e timer-wait ;

\ ----------------------------------------------------------------------
\ One wire communication with bq2010
\ ----------------------------------------------------------------------

1 pin-a bq-data

: bq-low ( -- ) bq-data low bq-data >output ;         \ Low output state

: bq-high ( -- ) bq-data >input ;                     \ Input state

\ Times come from the bq2010 datasheet
: tb ( -- ) $fd timer-wait ;                                     \ 3 delays
: tbr ( -- ) $ff timer-wait ;                                    \ 1 delay
: tsrh ( -- ) $ff timer-wait ;                                   \ 1 delay
:: tdv-loop ( n -- )                                             \ 256n delays
  tmp1 v-for 0 timer-wait tmp1 v-next ;
: tdv ( -- ) 5 tdv-loop ;                                        \ 1.576ms
: tsv ( -- ) 10 tdv-loop ;                                       \ 3.152ms
: tsrb ( -- ) 3 tdv-loop ;

: break ( -- ) bq-low tb bq-high tbr ;

: bit-one ( -- ) bq-high tdv ;
: bit-zero ( -- ) tdv bq-high ;

: send-lsb ( b -- b ' ) rrf-tos c bit-set? if bit-one exit then bit-zero ;

variable current-bit
: bq-write-byte ( b -- )
    8 current-bit
      v-for bq-low tsrh send-lsb tsv current-bit
    v-next
    drop
;

macro
: wait-for-low ( -- ) begin bq-data high? while repeat ;
: wait-for-high ( -- ) begin bq-data high? until ;
target

: add-lsb ( b -- b' )
    wait-for-low tsrb 2* bq-data high? if 1 + then wait-for-high
;

: bq-read-byte ( -- b ) 0 8 current-bit v-for add-lsb current-bit v-next ;

: bq-write ( content addr -- ) break bq-write-byte bq-write-byte ;
: bq-read ( addr -- content ) break bq-write-byte bq-read-byte ;

\ ----------------------------------------------------------------------
\ Version and information strings
\ ----------------------------------------------------------------------

variable vsn-count
variable vsn-src
variable vsn-dst

\ Copy 32 bytes long string from eeprom to i2c-buffer
:: i2c-buffer-copy ( src -- )
  vsn-src !
  i2c-buffer vsn-dst !
  $20 vsn-count v-for
    vsn-src @ ee@ vsn-dst @ ! 1 vsn-src +! 1 vsn-dst +!
  vsn-count v-next
;

: i2c-get-vsn ( -- )
  \ $Format: "  s\" SPIF-PIC $ProjectVersion$ (PicForth $ReleaseVersion$)  \""$
  s" SPIF-PIC 0.211 (PicForth 0.18)  "
  drop i2c-buffer-copy ;

: i2c-get-copyright ( -- )
  s" Copyright (c) 2002 ENST       "
  drop i2c-buffer-copy ;  

\ ----------------------------------------------------------------------
\ Event signalling
\ ----------------------------------------------------------------------

4 pin-a interrupt
variable event
1 constant startup
2 constant smartcard
4 constant i2c-error

:: signal-event ( n -- ) event or! interrupt toggle ;

\ ----------------------------------------------------------------------
\ Smartcard
\ ----------------------------------------------------------------------

\ The smartcard is access through a Philips TDA8004 device, which takes
\ care of proper power-up and power-down sequences.

0 pin-b i/o
1 pin-b clkdiv1
2 pin-b clkdiv2
3 pin-b aux1
4 pin-b aux2
5 pin-b /off

0 pin-c rstin
1 pin-c /cmdvcc

variable atr-length
create atr $10 allot

variable sc-prev
variable sc-count
variable sc-addr
variable sc-inverse
variable sc-parity
variable sc-max-length

\ The card can use up to 9600 bits (i.e. around one second) before
\ starting the next byte. Each time it is possible to deduce the right
\ size, it should be used to avoid useless waiting time.

: next-start-bit? ( -- f )
    60 begin
	dup while
	64 begin
	    dup while
	    0 tmr0 ! t0if bit-clr
	    begin
		i/o low? if 2drop true delay-half-bit exit then
	    t0if bit-set? until
	1- repeat drop
    1- repeat drop false
;

\ Smartcards can be operated in two modes: direct convention uses high-level
\ ones with LSB transmitted first, while inverse convention uses low-level
\ ones with MSB transmitted first. In both cases, even parity is used.

: get-bit-inverse ( b -- b' )
    1 lshift i/o high? if exit then 1 or 1 sc-parity xor! ;
: get-bit-direct ( b -- b' )
    1 rshift i/o low? if exit then 80 or 1 sc-parity xor! ;
: get-bit ( b -- b' )
    sc-inverse @ if get-bit-inverse exit then get-bit-direct ;

: get-byte ( -- b )
    0 sc-parity ! 0
    8 sc-count v-for delay-bit get-bit sc-count v-next
    \ Check parity
    delay-bit i/o low? if 1 sc-parity xor! then
    sc-parity @ 1 and if
	1 sc-parityerrors +!
	delay-bit i/o low i/o >output delay-bit delay-bit i/o >input
	drop next-start-bit? 0= if 0 exit then   \ Dead in the water...
	recurse exit
    then
    \ Skip first half of stop bit and record byte received
    delay-bit 1 sc-fromcard +!
;

: send-bit-inverse ( b -- b' )
    rlf-tos c bit-set? if i/o low 1 sc-parity xor! exit then i/o high ;
: send-bit-direct ( b -- b' )
    rrf-tos c bit-set? if i/o high exit then i/o low 1 sc-parity xor! ;
: send-bit ( b -- b' )
    sc-inverse @ if send-bit-inverse exit then send-bit-direct ;

: send-byte ( b -- )
    0 sc-parity !
    \ Preserve byte
    dup
    \ Start bit
    i/o low i/o >output delay-bit
    \ Send bits
    8 sc-count v-for send-bit delay-bit sc-count v-next drop
    \ Send parity
    sc-parity @ 1 and if i/o low else i/o high then delay-bit
    \ Send first stop bit
    i/o >input delay-bit
    \ Check NACK (parity error) from card
    i/o low? if 1 sc-nack +! delay-bit delay-bit recurse exit then
    \ Ok, remove copy and record byte
    drop 1 sc-tocard +!
;

: send-command ( length -- )
    begin
	sc-addr @ @ send-byte
	1- dup while
	1 sc-addr +! delay-bit
    repeat drop
;

: get-from-card ( addr max-length -- length )
    sc-max-length ! dup sc-addr !
    i/o >input next-start-bit? 0= if drop 0 exit then
    begin
	get-byte sc-addr @ ! 1 sc-addr +! 1 sc-max-length -!
	sc-max-length @ if next-start-bit? else false then while
    repeat
    sc-addr @ - negate
;

: smartcard-command ( -- )
    led-blink
    \ Send data to the card
    i2c-data 2 + sc-addr ! i2c-data @ send-command
    \ Get first byte or consider there is no answer
    i/o >input 0 \ Dummy byte for begin/repeat structure
    begin
	drop next-start-bit? 0= if
	   0 i2c-buffer ! smartcard signal-event led-on exit
        then
        get-byte
        dup $60 = while   \ 0x60 means that the card needs an extra time
    repeat
    dup i2c-data 3 + @ <> if
	\ Not a ACK (complement of INS), this is SW1, get SW2
	i2c-buffer 1+ ! i2c-buffer 2 + 1 get-from-card 1+ i2c-buffer !
    else
	\ Discard ack and get the answer plus SW1/SW2
	drop i2c-buffer 1+ i2c-data 1+ @ 2 + get-from-card i2c-buffer !
    then
    \ Signal that result is available
    smartcard signal-event led-on
;

: card-change? ( -- ) /off bit-set? sc-prev @ xor ;

: card-idle ( -- )
    i/o >input /cmdvcc high rstin low 0 sc-prev ! 0 atr-length ! ;

: handle-insertion ( -- )
    \ Debounce contact (do nothing if this is not a real card insertion)
    $20 sc-count v-for card-change? 0= if exit then sc-count v-next
    \ Record card presence
    led-blink /off bit-mask sc-prev !
    \ Use direct mode to read first ATR byte
    0 sc-inverse !
    \ Startup sequence (see TDA8004 datasheet)
    /cmdvcc low 140s rstin high
    \ Get first byte and check convention
    i/o >input next-start-bit? 0= if rstin low led-blink-fast exit then
    get-byte dup $03 = if drop $3f 1 sc-inverse +! then atr !
    \ Receive rest of ATR
    atr 1+ $10 get-from-card 1+ atr-length !
    \ Smartcard is ready
    smartcard signal-event led-on
;

\ While power to the card is removed, it looks like some capacitors
\ do still present the card as being here. A 80ms delay takes care of
\ that situation.

: handle-removal ( -- ) led-off card-idle 80ms smartcard signal-event ;
    
: handle-smartcard ( -- )
    sc-prev @ if handle-removal exit then handle-insertion ;

\ ----------------------------------------------------------------------
\ I2C
\ ----------------------------------------------------------------------

\ As explained in Microchip MSSP Module Silicon Errata Sheet, it is sometimes
\ necessary to reset the I2C module after a timeout because of a hardware
\ bug with no simple workaround.

variable i2c-timeout
: reset-i2c-module ( -- ) sspen bit-clr sspen bit-set ;

:: write-i2c ( b -- ) sspbuf ! ckp bit-set ;

: read-i2c ( -- b )
  sspbuf @
  sspov bit-clr? if exit then
  \ Record overflow condition
  sspov bit-clr 1 i2c-overflow +! ;

macro
: ack-i2c ( -- ) sspif bit-clr ;
target

: i2c-bq-write ( -- ) i2c-data 1+ @ i2c-data @ bq-write ;
: i2c-bq-read ( -- ) i2c-data @ bq-read i2c-buffer ! ;

: i2c-get-stats ( -- )
    i2c-overflow @ i2c-buffer !
    sc-parityerrors @ i2c-buffer 1+ ! sc-nack @ i2c-buffer 2 + !
    sc-tocard @ i2c-buffer 3 + ! sc-fromcard @ i2c-buffer 4 + !
;

: i2c-configure-pwm ( -- )
    i2c-data @ pr2 !
    i2c-data 1+ @ ccpr1l !
    i2c-data 2 + @ 2 rshift ccp1con @ $30 /and or ccp1con !
    i2c-data 3 + @ t2con @ 3 /and or t2con !
;

: i2c-start-pwm ( -- )
    portc 2 >output
    tmr2on bit-set
;

: i2c-stop-pwm ( -- )
    tmr2on bit-clr
    portc 2 >input
;

: i2c-get-event ( -- ) event @ i2c-buffer ! 0 event ! ;

: i2c-set-address ( -- ) i2c-data @ i2c-address ee! ;

: i2c-get-version ( -- ) vsn-major i2c-buffer ! vsn-minor i2c-buffer 1+ ! ;

: i2c-get-atr ( -- ) atr i2c-buffer atr-length @ cmove ;

\ There has been an incoming request, handle it
: i2c-incoming ( -- )
    \ Parameterless commands
    i2c-length @ 1 = if
      \ 0x01: get interface version information
      i2c-command @ $01 = if i2c-get-version exit then
      \ 0x04: get the stack pointer value
      i2c-command @ $04 = if fsr @ i2c-buffer ! exit then
      \ 0x06: call monitor (stop interrupts first)
      i2c-command @ $06 = if disable-interrupts ]asm 600 goto asm[ then
      \ 0x20: get statistics
      i2c-command @ $20 = if i2c-get-stats exit then
      \ 0x21: reset statistics
      i2c-command @ $21 = if reset-stats exit then
      \ 0x22: get version string
      i2c-command @ $22 = if i2c-get-vsn exit then
      \ 0x23: get copyright
      i2c-command @ $23 = if i2c-get-copyright exit then
      \ 0x31: start PWM
      i2c-command @ $31 = if i2c-start-pwm exit then
      \ 0x32: stop PWM
      i2c-command @ $32 = if i2c-stop-pwm exit then
      \ 0x40: get event
      i2c-command @ $40 = if i2c-get-event exit then
      \ 0x50: reset device
      i2c-command @ $50 = if reset then
      \ 0x60: get ATR length
      i2c-command @ $60 = if atr-length @ i2c-buffer ! exit then
      \ 0x61: get ATR
      i2c-command @ $61 = if i2c-get-atr exit then
      \ 0x63: reset smartcard
      i2c-command @ $63 = if handle-removal exit then
    then
    \ Commands with one parameter
    i2c-length @ 2 = if
      \ 0x02: read one byte in bank 0 (addr)
      i2c-command @ $02 = if i2c-data @ @ i2c-buffer ! exit then
      \ 0x11: bq2010 read (addr)
      i2c-command @ $11 = if i2c-bq-read exit then
      \ 0x51: set i2c address
      i2c-command @ $51 = if i2c-set-address exit then
    then
    \ Commands with two parameters
    i2c-length @ 3 = if
      \ 0x10: bq2010 write (addr, content)
      i2c-command @ $10 = if i2c-bq-write exit then
    then
    \ Commands with four parameters
    i2c-length @ 5 = if
      \ 0x30: configure PWM
      i2c-command @ $30 = if i2c-configure-pwm exit then
    then
    \ Commands with variable number of parameters
    i2c-length @ 2 < if exit then
    i2c-command @ $62 = if
      \ 0x62: send command to smartcard
      i2c-length @ i2c-data @ - 3 = if smartcard-command exit then
    then
    \ Command is unknown or only partially available
;

: reset-i2c-buffer ( -- ) i2c-buffer i2c-addr ! ;

\ State 1: make a dummy read and reset reception area
: i2c-state1 ( -- )
    80 i2c-timeout !   \ Arm timer
    read-i2c drop 0 i2c-length ! reset-i2c-buffer ;

\ State 2: add byte to command
: i2c-state2 ( -- )
    40 i2c-timeout !   \ Arm timer
    read-i2c i2c-addr @ ! 1 i2c-addr +! 1 i2c-length +! i2c-incoming ;

\ State 4: send next byte
: i2c-state4 ( -- ) i2c-addr @ @ write-i2c 1 i2c-addr +! ;

\ State 3: reset reception area and send first byte
: i2c-state3 ( -- ) reset-i2c-buffer i2c-state4 ;

\ State 5: end of read, nothing to do
: i2c-state5 ( -- ) ;

: handle-i2c ( -- )
    \ Acknowledge i2c and reset timeout
    ack-i2c
    \ Mask out unimportant bits
    sspstat @ $2d and
    \ Write operation, last byte was an address, buffer full
    dup $09 = if drop i2c-state1 exit then
    \ Write operation, last byte was data, buffer full
    dup $29 = if drop i2c-state2 exit then
    \ Read operation, last byte was an address, buffer empty
    dup $0c = if drop i2c-state3 exit then
    \ Read operation, last byte was data, buffer empty
    dup $2c = if drop i2c-state4 exit then
    \ A NACK has been received by the master, logic is reset
    $28 = if i2c-state5 exit then
    \ This is an inconsistent state, signal an error
    i2c-error signal-event ;

\ ----------------------------------------------------------------------
\ Interrupt handler
\ ----------------------------------------------------------------------

isr
: isr-handler ( -- )
    tmr1if bit-clr
    blinking-speed @ 1 and if
	1 blinking-counter -!
	z bit-set? if
	    led toggle blinking-speed @ blinking-counter !
	then
    then
    1 i2c-timeout -! z bit-set? if reset-i2c-module then
    isr-restore-return ;

\ ----------------------------------------------------------------------
\ Initialization and main program
\ ----------------------------------------------------------------------

: init ( -- )
    \ Timer0 is used with a prescaler of 8
    $02 option_reg !
    $06 adcon1 !          \ Disable A/C converter
    startup event !       \ Nothing happened yet
    interrupt >output     \ Interrupt line is an output
    \ PWM output is used as a host-controlled baud-rate generator
    ccp1con @ $0f /and $0c or ccp1con !   \ Activate PWM mode
    \ I2C slave mode is used
    $36 sspcon !          \ Enable I2C slave 7bit add., do not hold clock
    i2c-address ee@ 2* sspadd ! \ Set I2C address
    0 sspstat !           \ Clear all error condition bits
    0 sspcon2 !           \ Do not answer general call
    0 i2c-timeout !       \ Cf. variable declaration
    \ Ports b1, b2, c0 and c1 are outputs
    $f9 trisb !
    $fc trisc !
    clkdiv1 low clkdiv2 low   \ Divide frequency by 8
    card-idle              \ By default, do as if card is not present
    \ Statistics counter
    reset-stats
    \ Turn led off
    led >output led-off
    \ Timer 1
    $31 t1con ! tmr1ie bit-set peie bit-set enable-interrupts
;

main
: main ( -- )
    init
    begin
	sspif bit-set? if handle-i2c then
	card-change? if handle-smartcard then
    again ;

\ Configuration word

fosc-hs set-fosc      \ High-speed (26MHz) oscillator
false set-wdte        \ Do not use watchdog timer
false set-lvp         \ No low-voltage programming
false set-boden       \ No brown-out reset
true set-wrt          \ Enable flash-writing (monitor)
false set-/pwrte      \ Power-up timer
